from typing import Callable
import jax 
import jax.numpy as jnp
from jax._src.basearray import Array
from jax.tree_util import Partial
from jax._src.prng import PRNGKeyArray

# Economic parameters used in functions
T = 10 # number of periods: t=0, ...,T where death occurs at T
beta = 0.98 # discount factor
gamma = 1.1 # relative risk aversion
R = 1/beta # interest rate
rho = 0.9 # persistence of AR(1)
sigma_y = 0.2 # transitory shock variable
sigma_y_st = sigma_y / jnp.sqrt(1 - rho**2) # stationary variable of AR(1)
logy_bound = [-2 * sigma_y_st, 2 * sigma_y_st] # bound for log income
a_bound = [0, 10] # bound for assets

@jax.jit
def u(state:Array, action:Array) -> Array:
    '''
    Reward function
    '''
    return (beta**state[:,[0]])*(action ** (1. - gamma))/(1. - gamma)

@jax.jit
def m(key:PRNGKeyArray, state:Array, action:Array) -> Array:
    '''
    State evolution equation
    '''
    N = state.shape[0]
    t, y, a = state[:,[0]], state[:,[1]], state[:,[2]]
    c = action
    t_next = t + 1
    y_next = jnp.exp(rho * jnp.log(y) + sigma_y * jax.random.normal(key, shape = (N,1)))
    a_next = R * (a + y - c)
    return jnp.column_stack([t_next, y_next, a_next])

@jax.jit
def Gamma(state:Array) -> list[tuple[Array,Array]]:
    '''
    Define bounds of action in each state
    '''
    return [(jnp.ones((state.shape[0],1))*1e-6, state[:,[1]]+state[:,[2]])]

@Partial(jax.jit,static_argnames='N')
def F(key:PRNGKeyArray, N:int) -> Array:
    '''
    Sample N initial states
    '''
    t = jnp.zeros(N)
    key, subkey = jax.random.split(key)
    y = jnp.exp(jax.random.uniform(subkey, shape = (N,1), minval = logy_bound[0], maxval = logy_bound[1]))
    key, subkey = jax.random.split(key)
    a = jax.random.uniform(subkey, shape = (N,1), minval = a_bound[0], maxval = a_bound[1])
    state = jnp.column_stack([t, y, a])
    return state

@jax.jit
def nn_to_action(state:Array, 
                 params:dict,
                 nn:Callable[[dict, Array], Array]
                 ) -> Array:
    '''
    Defines how a Haiku Neural Network, nn, with parameters, params, is mapped
    into an action.
    
    Parameters:
    -----------
    state: current state = N_simul x n_states
    nn: Haiku Neural Network with signature nn(params, state)
    params: dictionary of parameters used by nn. 

    Returns:
    -----------
    action: action to take = N_simul x n_actions.
    '''
    c_min, c_max = Gamma(state)[0]
    action = c_min + nn(params, state) * c_max
    return action



import jax
import jaxlib
import jax.numpy as jnp
from jax._src.basearray import Array
from jax.tree_util import Partial
import haiku as hk
from typing import Callable

def policy_nn(X:Array,
              P:int, 
              N_nodes:int,
              N_hidden:int,
              f_activation:Callable,
              f_outputs:list
              ) -> Array:
    '''
    This function defines a policy function that maps a matrix of states (exogeneous and endogeneous) into the 
    model policy functions. The policy function is a neural network. 

    Parameters:
    ----------
    X: N x K matrix of states, where K is the number of states and the function is vectorized over the N rows
    P: dimension of output layer = dimension of policy function
    N_nodes: number of nodes in each layer except output layer
    N_layers: number of hidden layers
    f_activation: activation function use in all but output layers (e.g. jax.nn.tanh, jax.nn.relu)
    f_outputs: list of Callable jaxlib.xla_extension.CompiledFunction (e.g. jax.nn.sigmoid). 
        The length of this list must be equal to P, as you need a separate activation function for each output. 
        
    Returns:
    ----------
    N x P matrix of policies, where P is dimension of policy function
    '''
    # Input layer + hidden laters
    for i in range(N_hidden + 1):
        X = hk.Linear(N_nodes)(X)
        X = f_activation(X)
    # Output layer
    X = hk.Linear(P)(X)
    Y = [f_outputs[p](X[:, p]) for p in range(P)]        
    return jnp.column_stack(Y)

def initialize_nn(key:Array,
                  K:int, 
                  P:int, 
                  N_nodes:int,
                  N_hidden:int,
                  f_activation:jaxlib.xla_extension.CompiledFunction,
                  f_outputs:list
                  ) -> list[dict, Callable]:
    '''
    This function initializes a Haiku Neural Network policy function with the given set of parameters. 
    
    Parameters:
    ----------
    key: JAX RNG key
    K: number of state variables in model
    P: dimension of output layer = dimension of policy function
    N_nodes: number of nodes in each layer except output layer
    N_layers: number of hidden layers
    f_activation: activation function use in all but output layers (e.g. jax.nn.tanh, jax.nn.relu)
    f_outputs: list of Callable jaxlib.xla_extension.CompiledFunction (e.g. jax.nn.sigmoid). 
        The length of this list must be equal to P, as you need a separate activation function for each output. 
        
    Returns:
    ----------
    params: initialized parameters of NN
    nn: policy function that is a neural network
    '''  
    # HK transform to get policy function that takes parameters as inputs
    init, policy_args = hk.without_apply_rng(hk.transform(policy_nn))
    # Initialize parameters
    x0 = jnp.column_stack([0.]*K)
    key, subkey = jax.random.split(key)
    params = init(subkey, x0, P, N_nodes, N_hidden, f_activation, f_outputs)
    # Make a function that surpress NN parameters
    def nn(p, X):
        return policy_args(p, X, P, N_nodes, N_hidden, f_activation, f_outputs)
    return params, nn

def make_policy_function(nn_to_action:Callable,
                         key:Array,
                         K:int, 
                         P:int, 
                         N_nodes:int,
                         N_hidden:int,
                         f_activation:jaxlib.xla_extension.CompiledFunction,
                         f_outputs:list
                         ) -> Callable:
    '''
    This function creates a policy function for the model of interest that is Haiku Neural
    Network with the supplied set of parameters.
    
    Parameters:
    ----------
    nn_to_action: user-defined function that characterizes how outputs of neural network
        translate into an action in the model
    key: JAX RNG key
    K: number of state variables in model
    P: dimension of output layer = dimension of policy function
    N_nodes: number of nodes in each layer except output layer
    N_layers: number of hidden layers
    f_activation: activation function use in all but output layers (e.g. jax.nn.tanh, jax.nn.relu)
    f_outputs: list of Callable jaxlib.xla_extension.CompiledFunction (e.g. jax.nn.sigmoid). 
        The length of this list must be equal to P, as you need a separate activation function for each output. 
        
    Returns:
    ----------
    params: initialized parameters of NN
    policy: policy function that is a neural network with signature policy(state, params)
    '''
    params, nn = initialize_nn(key = key,
                               K = K,
                               P = P,
                               N_nodes = N_nodes,
                               N_hidden = N_hidden,
                               f_activation = f_activation,
                               f_outputs = f_outputs
                               )
    policy = Partial(nn_to_action, nn = Partial(nn))
    return params, policy


import jax
from jax._src.prng import PRNGKeyArray
from jax._src.basearray import Array
from typing import Callable
import jax.numpy as jnp
from jax.tree_util import Partial
import optax

def evaluate_policy(key:PRNGKeyArray,
                    policy:Callable[[Array, Array], Array],
                    u:Callable[[Array, Array], Array],
                    m:Callable[[PRNGKeyArray, Array, Array], Array], 
                    s0:Array,
                    T:int,
                    N_simul:int=1
                    ) -> Array:
    '''
    Calculates value function for a given set of initial states, simulating
    N_simul paths for each k=1,...,K initial states.
    
    Parameters:
    ----------
    key: JAX random number key
    policy: policy function
    u: instantaneous utility function
    m: law of motion
    s0: K x n_states array, where K is different initial states to consider
    T: number of periods
    N_simul: number of simulations to be considered to compute expectation
    
    Returns:
    ----------
    Vector of value functions at s0(k) = 
        E(\sum_{t=0}^T \beta^t u(action_t, state_t) | state_0 = s0(k))
    '''
    K = s0.shape[0]
    state = jnp.repeat(s0, repeats = N_simul, axis = 0)
    V = jnp.zeros((K * N_simul, 1))
    key, *subkey = jax.random.split(key, (T + 1))
    for t in range(T):
        action = policy(state)
        V += u(state, action) 
        state = m(subkey[t], state, action)
    V = V.reshape(K, -1)
    return jnp.mean(V, axis = 1, keepdims = True)                    

def train_step(key:PRNGKeyArray,
               params:dict,
               policy:Callable[[Array, Array], Array],
               u:Callable[[Array, Array], Array],
               m:Callable[[PRNGKeyArray, Array, Array], Array], 
               s0:Array,
               T:int,
               N_simul:int,
               optimizer,
               opt_state
               ):
    '''
    Performs one optimization step to update neural network parameters, params,
    where objective function is -1 times the equal-weighted mean of value
    functions across K different initial states.
    
    Parameters:
    ----------
    key: JAX random number key
    params: dictionary of neural network parameters
    u: Instantaneous utility function
    m: Law of motion
    s0: Initial state = K x n_states array, where K is different
        initial states to consider
    T: number of periods
    N_simul: number of simulations to compute value function for each s0
    optimizer: optax optimizer object
    opt_state: state of optax optimizer
    
    Returns:
    ----------
    params: updated neutral network parameters
    opt_state: updated state of optimizer
    value: objective function value at previous step
    '''
    value, grads = jax.value_and_grad(lambda params: -jnp.mean(
        evaluate_policy(key = key, 
                       policy = Partial(policy, params = params),
                       u = u,
                       m = m,
                       s0 = s0,
                       T = T,
                       N_simul = N_simul)
        )
    )(params)
    updates, opt_state = optimizer.update(grads, opt_state)
    params = optax.apply_updates(params, updates)
    return params, opt_state, value

def train(key:PRNGKeyArray,
          params:dict,
          policy:Callable[[Array, Array], Array],
          u:Callable[[Array, Array], Array],
          m:Callable[[PRNGKeyArray, Array, Array], Array], 
          F:Callable[[Array, Array], Array],
          T:int,
          N_simul:int,
          batch_size:int,
          N_iter:int,
          optimizer
          ):
    '''
    Optimizes parameters of a neural network, policy, by performing multiple steps
    of stochastic gradient descent. 
    
    Parameters:
    ----------
    key: JAX random number key
    params: dictionary of initialized neural network parameters
    policy: neural network to optimize
    u: Instantaneous utility function
    m: Law of motion
    F: function to initialize states
    T: number of periods
    N_simul: number of simulations to compute value function for each s0
    batch_size: number of initial states, s0, to sample in each optimization step
    N_iter: number of optimization steps
    optimizer: optax optimizer object
    
    Returns:
    ----------
    params: neural network parameters at final step of optimization
    '''
    opt_state = optimizer.init(params)
    for i in range(N_iter):
        key, subkey = jax.random.split(key)
        s0 = F(key = subkey, N = batch_size)
        key, subkey = jax.random.split(key)
        params, opt_state, value = train_step(key = subkey, 
                                              params = params, 
                                              policy = policy, 
                                              u = u, 
                                              m = m, 
                                              s0 = s0, 
                                              T = T, 
                                              N_simul = N_simul, 
                                              optimizer = optimizer, 
                                              opt_state = opt_state)
        print(f'\rObjective value on training iteration {i} out of {N_iter}: {-value}', end='')
    return params

def simulate(key:Array,
             policy:Callable[[Array, Array], Array],
             params:dict,
             u:Callable[[Array, Array], Array],
             m:Callable[[PRNGKeyArray, Array, Array], Array], 
             s0:Array,
             T:int
             ) -> float:
    '''
    Simulates multiple paths given current policy function and
    save the results. This function is used after params have been
    optimized via train(). 

    Parameters:
    ----------
    key: JAX random number key
    policy: Policy function. Given a state, returns the action. 
    params: NN parameters
    u: Instantaneous utility function
    m: Law of motion
    s0: Initial state (N_simul x n_states)
    T: number of periods
    
    Returns:
    ----------
    array of simulation results, which comes from stacking the following
        intermediate arrays
            states: jnp.ndarray
                states in simulations (N_simul x N_states x T)
            actions: jnp.ndarray
                policy functions in simulations (N_simul x T)
            values: jnp.ndnarray
                value function in simulations (N_simul x T)
    '''
    N_simul, N_states = s0.shape
    states = jnp.zeros((N_simul, N_states, T))
    states = states.at[:,:,0].set(s0)
    actions = jnp.zeros((N_simul, T))
    values = jnp.zeros((N_simul, T))
    for t in range(T):
        key, subkey = jax.random.split(key)
        state = states[:,:,t]
        action = policy(state, params)
        actions = actions.at[:,[t]].set(action)
        values = values.at[:,:t+1].set(u(state, action) + values[:,:t+1])
        if t + 1 < T:
            states = states.at[:,:,t+1].set(m(subkey, state, action))
    outputs = [jnp.repeat(jnp.array(range(1, N_simul + 1)), T)]
    for k in range(N_states):
        outputs.append(states[:,k,:].reshape(-1,1))
    outputs.append(actions.reshape(-1,1))
    outputs.append(values.reshape(-1,1))
    return jnp.column_stack(outputs)


from setuptools import setup


with open("README.md","r") as fh:
    long_description = fh.read()

setup(
    name = 'nndp',
    version = '0.0.5',
    description = 'Dynamic Programming with Neural Networks',
    long_description = long_description,
    long_description_content_type='text/markdown',
    py_modules = ["nndp.policy_function", "nndp.core"],
    package_dir={'':'src'},
    author='Marc de la Barrera i Bardalet',
    url = 'https://github.com/marcdelabarrera/nndp',
    author_email='mbarrera@mit.edu',
    install_requires = ["numpy >= 1.20.0","jax >= 0.3.22", "optax >= 0.1.0", "jaxlib >= 0.3.22", "dm-haiku >= 0.0.9"],
    extras_require={"dev":["pytest >= 7.1.2",]},
    classifiers =[
        "Programming Language :: Python :: 3.10"
    ]
)